/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "pic32_c.h"
#include "init.h"
#include "debug.h"

/* Specific functionality. */

#include "devconfig.h"
#include "display.h"
#include "display_config.h"
#include "font.h"
#include "main.h"
#include "vga.h"
#include "vga_display.h"



/* Define DMA channels if not indicated in the build configuration. */

/* CPU-based transfers: no channels. */

#ifdef TRANSFER_CPU
#define LINE_CHANNELS   0
#define SCROLL_XSTEP    1

/* DMA-based transfers: single channel by default. */

#else

#ifndef LINE_CHANNELS
#define LINE_CHANNELS   1
#endif

#define SCROLL_XSTEP    LINE_CHANNELS
#endif



/* Define timers if not indicated in the build configuration. */

#ifndef LINE_TIMER
#define LINE_TIMER      2
#endif

#ifndef TRANSFER_TIMER
#define TRANSFER_TIMER  0
#endif

/* Define different output ports and pins for parallel mode. */

#ifndef PARALLEL_MODE
#define VGA_OUTPUT      PORTB
#define OC1_PIN         RPA0R
#define OC2_PIN         RPA1R
#define LED_PIN         (1 << 3)
#else
#define VGA_OUTPUT      PM_REG(0, PMxDIN)
#define OC1_PIN         RPB4R
#define OC2_PIN         RPB5R
#define LED_PIN         (1 << 2)
#endif



/* Define the relationship between source images and the screen. */

#ifndef SOURCE_HEIGHT
#define SOURCE_HEIGHT   256
#endif

#define SOURCE_YSTEP    (SOURCE_HEIGHT / LINE_COUNT)



/* Framebuffer memory. */

static uint8_t framebuffer[FRAME_SIZE * FRAME_COUNT];



/* Bundled image and font data. */

extern uint8_t screendata[];
extern uint32_t screendata_width, screendata_height;
extern uint8_t sprite[];
extern uint32_t sprite_width, sprite_height;

extern uint8_t fontchars[];
extern uint32_t fonttable[];
extern uint32_t fontbase, fontlimit;

static font_config_t font_config;



/* Busy wait. */

static void wait(uint32_t delay)
{
    uint32_t counter = delay;

    if (!delay) return;
    while (counter--) __asm__("");	/* retain loop */
}

/* Blink an attached LED with delays implemented using a loop. */

static void blink(uint32_t delay, uint32_t port, uint32_t pins)
{
    /* Clear outputs (LED). */

    CLR_REG(port, pins);

    while (1)
    {
        wait(delay);

        /* Invert outputs (LED). */

        INV_REG(port, pins);
    }
}

/* Copy to the store from the display, then blit the image. */

static void plot_sprite(uint8_t *background, int x, int y)
{
    copy_display(&display_config, background,
                 sprite_width, sprite_height / SOURCE_YSTEP, 1,
                 x, y, -1, 0);
    copy_display(&display_config, sprite,
                 sprite_width, sprite_height, SOURCE_YSTEP,
                 x, y, 0x8c, 1);
}

/* Copy to the display from the store, restoring the original background. */

static void unplot_sprite(uint8_t *background, int x, int y)
{
    copy_display(&display_config, background,
                 sprite_width, sprite_height / SOURCE_YSTEP, 1,
                 x, y, -1, 1);
}

/* Plot the revealed region at the edge of the screen after scrolling. */

static void plot_screen_edge(int xorigin, int yorigin, int xstep)
{
    /* The display region is either the left or right edge. */

    int xdisplay = xstep < 0 ? 0 : screendata_width - xstep;

    /* The source region depends on the origin within the background image. */

    int xsource = (xdisplay + xorigin) % screendata_width;

    /* The column width is the absolute increment. */

    int width = xstep < 0 ? -xstep : xstep;

    /* Plot a column in two pieces if the vertical origin is
       non-zero. The first piece is at (xdisplay, 0) and
       provides the lower part of the background image displaced
       upwards (or downwards having wrapped around) on the
       screen. */

    copy_display_section(&display_config, screendata,
                         screendata_width, screendata_height,
                         xsource, yorigin,
                         width, screendata_height - yorigin,
                         SOURCE_YSTEP,
                         xdisplay, 0,
                         -1, 1);

    /* The second column is at (xdisplay, h - yorigin) and
       provides the upper part of the background image displaced
       downwards (or upwards having wrapped around) on the
       screen. */

    if (yorigin)
        copy_display_section(&display_config, screendata,
                             screendata_width, screendata_height,
                             xsource, 0,
                             width, yorigin,
                             SOURCE_YSTEP,
                             xdisplay, (screendata_height - yorigin) / SOURCE_YSTEP,
                             -1, 1);
}

/* Wrap a value within the bounds [0, limit) provided value is already within
   the bounds [-limit, limit * 2). */

static int wrap_value(int value, int limit)
{
    if (value < 0)
        return value + limit;
    else if (value >= limit)
        return value - limit;
    else
        return value;
}

/* Move a sprite around on the framebuffer. */

static void animate(uint32_t delay)
{
    /* Stored region behind the sprite. */

    uint8_t background[FRAME_COUNT][(sprite_width * sprite_height) / SOURCE_YSTEP];

    /* Positions of the stored regions for each frame. */

    int background_x[FRAME_COUNT], background_y[FRAME_COUNT];
    int background_stored[FRAME_COUNT];

    /* Screen start values for each frame. */

    uint32_t frame_offset[FRAME_COUNT];

    /* Sprite position. */

    int x, y;

    /* Scrolling directions. */

    int dir[] = {1, 0, -1, 0, 1};
    int dirindex = 0;

    /* Scrolling position. */

    int xorigin = 0, yorigin = 0;

    /* Scrolling positions for each frame. */

    int xorigins[FRAME_COUNT], yorigins[FRAME_COUNT];

    /* Scroll increments and replotted column details. */

    int xdir, ydir, xstep, ystep;

    /* Current frame being accessed. */

    int frame;

    for (frame = 0; frame < display_config.frames; frame++)
    {
        background_stored[frame] = 0;
        frame_offset[frame] = 0;
        xorigins[frame] = xorigin;
        yorigins[frame] = yorigin;
    }

    frame = display_config.frame;

    /* Animation loop. */

    while (1)
    {
        for (y = 0; y < display_config.line_count - sprite_height; y++)
        {
            for (x = 0; x < display_config.line_length - sprite_width; x++)
            {
                plot_sprite(background[frame], x, y);

                /* Record the stored background. */

                background_x[frame] = x;
                background_y[frame] = y;
                background_stored[frame] = 1;

                /* Update the display with the frame details. */

                vga_set_frame(&display_config);
                wait(delay);

                /* Select the next frame to plot to. */

                frame = wrap_value(frame + 1, display_config.frames);
                select_frame(&display_config, frame, frame_offset[frame]);

                /* Prepare the frame for updates. */

                if (background_stored[frame])
                    unplot_sprite(background[frame], background_x[frame], background_y[frame]);

                /* Scroll in the indicated direction. */

                xdir = dir[dirindex];
                ydir = dir[dirindex + 1];

                /* Update the origin if appropriate. */

                /* Due to the effect of a simple screen start increment in the
                   dual channel configuration, horizontal scrolling involves two
                   pixel increments and thus requires a two-pixel column to be
                   plotted. */

                if (xdir)
                    xorigin += xdir * SCROLL_XSTEP;

                if (ydir)
                    yorigin += ydir * SOURCE_YSTEP;

                /* Determine the magnitude of the scrolling required from this
                   frame to the current origin. */

                xstep = xorigin - xorigins[frame];
                ystep = yorigin - yorigins[frame];

                /* Scroll the frame. */

                scroll_display(&display_config, xstep / SCROLL_XSTEP,
                                                ystep / SOURCE_YSTEP);

                /* Record the new screen start offset. */

                frame_offset[frame] = get_start_offset(&display_config);

                /* Update the current origin. */

                xorigin = wrap_value(xorigin, screendata_width);
                yorigin = wrap_value(yorigin, screendata_height);

                /* For horizontal scrolling, plot the exposed column at the left
                   (if scrolling left) or at the right (if scrolling right). */

                if (xstep)
                    plot_screen_edge(xorigin, yorigin, xstep);

                /* Record the origin for this frame. */

                xorigins[frame] = xorigin;
                yorigins[frame] = yorigin;
            }

            /* Switch direction periodically. */

            dirindex = wrap_value(dirindex + 1, 4);
        }
    }
}

/* Fill the screen with characters. */

static void write_chars(void)
{
    int x = 0, y = 0;
    char c;

    font_config.chars = (char_definition_t *) fontchars;
    font_config.table = fonttable;
    font_config.base = fontbase;
    font_config.limit = fontlimit;

    while (y < display_config.line_count)
        for (c = (char) font_config.base; c < (char) font_config.limit; c++)
        {
            x = write_char(&display_config, &font_config, c, x, y, 0xff);

            if (x > display_config.line_length)
            {
                x = 0; y += 9;
            }
        }
}

/* Set up a background. */

static void setup(void)
{
    int frame;

    for (frame = 0; frame < display_config.frames; frame++)
    {
        /* Obtain the frame. */

        select_frame(&display_config, frame, 0);

        /* Plot the image centred on the screen. */

        copy_display(&display_config, screendata,
                     screendata_width, screendata_height,
                     SOURCE_YSTEP,
                     (display_config.line_length - screendata_width) / 2,
                     (display_config.line_count - (screendata_height / SOURCE_YSTEP)) / 2,
                     -1, 1);

        /* Write a sequence of characters. */

        write_chars();
    }

    select_frame(&display_config, 0, 0);
}



/* Main program. */

void main(void)
{
    init_memory();
    init_pins();
    init_outputs();

    unlock_config();
    config_oc();
    config_uart();
    lock_config();

#ifndef TRANSFER_CPU
    init_dma();
#endif

#ifdef PARALLEL_MODE
    init_pm();

    /* Configure parallel master mode. */

    pm_init(0, 0b10);
    pm_set_output(0, 1, 0);
    pm_on(0);
#endif

    uart_init(1, FPB, 115200);
    uart_on(1);

    /* Initialise memory for a double-buffered display. */

    init_display(&display_config, framebuffer, LINE_LENGTH, LINE_COUNT, FRAME_COUNT);

    /* Initialise VGA output with one or two line channels, configuring a line
       timer and any transfer timer, with an initiating channel being introduced
       if a transfer timer is specified. */

    init_vga_with_timers(&display_config, LINE_CHANNELS, LINE_TIMER, TRANSFER_TIMER);

    /* Configure VGA output transfer to the output register, also configuring
       output compare units for horizontal and vertical sync. */

    vga_configure_transfer(VGA_OUTPUT);
    vga_configure_sync(1, 2);

    interrupts_on();

    /* Move a sprite around on the screen with a delay between each movement. */

    setup();
    animate(1 << 18);
}



/* Exception and interrupt handlers. */

void exception_handler(void)
{
    blink(1 << 22, PORTA, LED_PIN);
}

void interrupt_handler(void)
{
    vga_interrupt_handler();
}



/* Peripheral pin configuration. */

void config_oc(void)
{
    /* Map OC1. */

    REG(OC1_PIN) = 0b0101;

    /* Map OC2. */

    REG(OC2_PIN) = 0b0101;
}

void config_uart(void)
{
    /* Map U1RX to RPB13. */

    REG(U1RXR) = 0b0011;        /* U1RXR<3:0> = 0011 (RPB13) */

    /* Map U1TX to RPB15. */

    REG(RPB15R) = 0b0001;       /* RPB15R<3:0> = 0001 (U1TX) */

    /* Set RPB13 to input. */

    SET_REG(TRISB, 1 << 13);
}
